@@ -25,7 +25,7 @@
 #include <QString>
 #include <QStringList>
 #include <QFile>
-#include <QRegExp>
+#include <QRegularExpression>
 #include <QPluginLoader>
 #include <QVariant>
 #include <QCoreApplication>
@@ -45,7 +45,7 @@ static void loadEmoticonsPlugin()
         // Check if QGuiApplication::platformName property exists. This is a
         // hackish way of determining whether we are running QGuiApplication,
         // because we cannot load the FrameworkIntegration plugin into a
-        // QCoreApplication, as it would crash immediatelly
+        // QCoreApplication, as it would crash immediately
         if (qApp->metaObject()->indexOfProperty("platformName") > -1) {
             QPluginLoader lib(QStringLiteral("kf5/KEmoticonsIntegrationPlugin"));
             QObject *rootObj = lib.instance();
@@ -143,6 +143,72 @@ QString KTextToHTMLHelper::getEmailAddress()
     return address;
 }
 
+QString KTextToHTMLHelper::getPhoneNumber()
+{
+    if (!mText[mPos].isDigit() && mText[mPos] != QLatin1Char('+')) {
+        return {};
+    }
+
+    static const QString allowedBeginSeparators = QStringLiteral(" \r\t\n:");
+    if (mPos > 0 && !allowedBeginSeparators.contains(mText[mPos - 1])) {
+        return {};
+    }
+
+    // this isn't 100% accurate, we filter stuff below that is too hard to capture with a regexp
+    static const QRegularExpression telPattern(QStringLiteral(R"([+0](( |( ?[/-] ?)?)\(?\d+\)?+){6,30})"));
+    const auto match = telPattern.match(mText, mPos, QRegularExpression::NormalMatch, QRegularExpression::AnchoredMatchOption);
+    if (match.hasMatch()) {
+        auto m = match.captured();
+        // check for maximum number of digits (15), see https://en.wikipedia.org/wiki/Telephone_numbering_plan
+        if (std::count_if(m.begin(), m.end(), [](const QChar &c) { return c.isDigit(); }) > 15) {
+            return {};
+        }
+        // only one / is allowed, otherwise we trigger on dates
+        if (std::count(m.begin(), m.end(), QLatin1Char('/')) > 1) {
+            return {};
+        }
+
+        // parenthesis need to be balanced, and must not be nested
+        int openIdx = -1;
+        for (int i = 0; i < m.size(); ++i) {
+            if ((m[i] == QLatin1Char('(') && openIdx >= 0) || (m[i] == QLatin1Char(')') && openIdx < 0)) {
+                return {};
+            }
+            if (m[i] == QLatin1Char('(')) {
+                openIdx = i;
+            } else if (m[i] == QLatin1Char(')')) {
+                openIdx = -1;
+            }
+        }
+        if (openIdx > 0) {
+            m = m.left(openIdx - 1).trimmed();
+        }
+
+        // check if there's a plausible separator at the end
+        static const QString allowedEndSeparators = QStringLiteral(" \r\t\n,.");
+        const auto l = m.size();
+        if (mText.size() > mPos + l && !allowedEndSeparators.contains(mText[mPos + l])) {
+            return {};
+        }
+
+        mPos += l - 1;
+        return m;
+    }
+    return {};
+}
+
+static QString normalizePhoneNumber(const QString &str)
+{
+    QString res;
+    res.reserve(str.size());
+    for (const auto c : str) {
+        if (c.isDigit() || c == QLatin1Char('+')) {
+            res.push_back(c);
+        }
+    }
+    return res;
+}
+
 bool KTextToHTMLHelper::atUrl() const
 {
     // the following characters are allowed in a dot-atom (RFC 2822):
@@ -170,7 +236,9 @@ bool KTextToHTMLHelper::atUrl() const
         (ch == QLatin1Char('w') && mText.midRef(mPos, 4) == QLatin1String("www.")) ||
         (ch == QLatin1Char('f') && (mText.midRef(mPos, 4) == QLatin1String("ftp.") ||
                                     mText.midRef(mPos, 7) == QLatin1String("file://"))) ||
-        (ch == QLatin1Char('n') && mText.midRef(mPos, 5) == QLatin1String("news:"));
+        (ch == QLatin1Char('n') && mText.midRef(mPos, 5) == QLatin1String("news:")) ||
+        (ch == QLatin1Char('t') && mText.midRef(mPos, 4) == QLatin1String("tel:"));
+
 }
 
 bool KTextToHTMLHelper::isEmptyUrl(const QString &url) const
@@ -188,7 +256,9 @@ bool KTextToHTMLHelper::isEmptyUrl(const QString &url) const
            url == QLatin1String("www") ||
            url == QLatin1String("ftp") ||
            url == QLatin1String("news") ||
-           url == QLatin1String("news://");
+           url == QLatin1String("news://") ||
+           url == QLatin1String("tel") ||
+           url == QLatin1String("tel:");
 }
 
 QString KTextToHTMLHelper::getUrl(bool *badurl)
@@ -224,7 +294,6 @@ QString KTextToHTMLHelper::getUrl(bool *badurl)
                 afterUrl = QLatin1Char('"');
             }
         }
-
         url.reserve(mMaxUrlLen);    // avoid allocs
         int start = mPos;
         bool previousCharIsSpace = false;
@@ -246,6 +315,16 @@ QString KTextToHTMLHelper::getUrl(bool *badurl)
                 }
                 mPos--;
             }
+            if (!previousCharIsSpace && (mText[mPos] == QLatin1Char(' ')) && ((mPos + 1) < mText.length())) {
+                // Fix kmail bug: allow "http://www.foo.bar http://foo.bar/"
+                // Therefore: check if what follows is another URL and if so, stop here
+                mPos++;
+                if (atUrl()) {
+                    mPos--;
+                    break;
+                }
+                mPos--;
+            }
             if (mText[mPos].isSpace()) {
                 previousCharIsSpace = true;
             } else if (!previousIsAnAnchor && mText[mPos] == QLatin1Char('[')) {
@@ -285,13 +364,14 @@ QString KTextToHTMLHelper::getUrl(bool *badurl)
         if (isEmptyUrl(url) || (url.length() > mMaxUrlLen)) {
             mPos = start;
             url.clear();
+            return url;
         } else {
             --mPos;
         }
     }
 
     // HACK: This is actually against the RFC. However, most people don't properly escape the URL in
-    //       their text with "" or <>. That leads to people writing an url, followed immediatley by
+    //       their text with "" or <>. That leads to people writing an url, followed immediately by
     //       a dot to finish the sentence. That would lead the parser to include the dot in the url,
     //       even though that is not wanted. So work around that here.
     //       Most real-life URLs hopefully don't end with dots or commas.
@@ -321,25 +401,28 @@ QString KTextToHTMLHelper::highlightedText()
         return QString();
     }
 
-    QRegExp re =
-        QRegExp(QStringLiteral("\\%1((\\w+)([\\s-']\\w+)*( ?[,.:\\?!;])?)\\%2").arg(ch).arg(ch));
-    re.setMinimal(true);
-    if (re.indexIn(mText, mPos) == mPos) {
-        int length = re.matchedLength();
-        // there must be a whitespace after the closing formating symbol
-        if (mPos + length < mText.length() && !mText[mPos + length].isSpace()) {
-            return QString();
-        }
-        mPos += length - 1;
-        switch (ch.toLatin1()) {
-        case '*':
-            return QLatin1String("<b>*") + re.cap(1) + QLatin1String("*</b>");
-        case '_':
-            return QLatin1String("<u>_") + re.cap(1) + QLatin1String("_</u>");
-        case '/':
-            return QLatin1String("<i>/") + re.cap(1) + QLatin1String("/</i>");
-        case '-':
-            return QLatin1String("<strike>-") + re.cap(1) + QLatin1String("-</strike>");
+    QRegularExpression re(QStringLiteral("\\%1([^\\s|^\\%1].*[^\\s|^\\%1])\\%1").arg(ch));
+    re.setPatternOptions(QRegularExpression::InvertedGreedinessOption);
+    const auto match = re.match(mText, mPos, QRegularExpression::NormalMatch, QRegularExpression::AnchoredMatchOption);
+
+    if (match.hasMatch()) {
+        if (match.capturedStart() == mPos) {
+            int length = match.capturedLength();
+            // there must be a whitespace after the closing formating symbol
+            if (mPos + length < mText.length() && !mText[mPos + length].isSpace()) {
+                return QString();
+            }
+            mPos += length - 1;
+            switch (ch.toLatin1()) {
+            case '*':
+                return QLatin1String("<b>*") + match.capturedRef(1) + QLatin1String("*</b>");
+            case '_':
+                return QLatin1String("<u>_") + match.capturedRef(1) + QLatin1String("_</u>");
+            case '/':
+                return QLatin1String("<i>/") + match.capturedRef(1) + QLatin1String("/</i>");
+            case '-':
+                return QLatin1String("<strike>-") + match.capturedRef(1) + QLatin1String("-</strike>");
+            }
         }
     }
     return QString();
@@ -469,7 +552,6 @@ QString KTextToHTML::convertToHtml(const QString &plainText, const KTextToHTML::
                     } else {
                         hyperlink = str;
                     }
-
                     result += QLatin1String("<a href=\"") + hyperlink + QLatin1String("\">") + str.toHtmlEscaped() + QLatin1String("</a>");
                     x += helper.mPos - start;
                     continue;
@@ -490,6 +572,14 @@ QString KTextToHTML::convertToHtml(const QString &plainText, const KTextToHTML::
                     x += str.length() - 1;
                     continue;
                 }
+                if (flags & ConvertPhoneNumbers) {
+                    str = helper.getPhoneNumber();
+                    if (!str.isEmpty()) {
+                        result += QLatin1String("<a href=\"tel:") + normalizePhoneNumber(str) + QLatin1String("\">") + str + QLatin1String("</a>");
+                        x += str.length() - 1;
+                        continue;
+                    }
+                }
             }
             if (flags & HighlightText) {
                 str = helper.highlightedText();
